/*
 * SPDX-FileCopyrightText: 2015-2024 CSSlayer <wengxt@gmail.com>
 *
 * SPDX-License-Identifier: LGPL-2.1-or-later
 *
 */
#include "eventlooptests.h"

int main() {
    runAllEventLoopTests();
    return 0;
}