/*
 * SPDX-FileCopyrightText: 2024-2024 CSSlayer <wengxt@gmail.com>
 *
 * SPDX-License-Identifier: LGPL-2.1-or-later
 *
 */

#ifndef _TEST_EVENTLOOPTESTS_H_
#define _TEST_EVENTLOOPTESTS_H_

void runAllEventLoopTests();

#endif